from __future__ import print_function
import numpy as np
from scipy.optimize import minimize as scipimin

def compute_saxs(q,params):
    """
    Given q and a dict of parameters,
    compute the saxs spectrum.
    Supported parameters are the same as 
    SaxsParameterization Operation outputs:
    I0: Intensity at q=0, by fitting a polynomial to the low-q region with dI/dq(q=0)=0. 
    r0_pre: precursor term radius (Angstrom). 
    I0_pre: precursor intensity scaling factor. 
    r0_sphere: mean radius (Angstrom) of sphere population. 
    sigma_sphere: standard deviation (Angstroms) of sphere population radii. 
    I0_sphere: spherical population intensity scaling factor. 
    q_pk0: q value (1/Angstrom) for location of fundamental (lowest-q) diffraction peak. 
    sigma_pk: width parameter for Pseudo-Voigt diffraction peak profile. 
    I0_pk: diffraction intensity scaling factor.

    TODO: Document the equation.
    """
    if params.keys() == ['I_at_0']:
        return params['I_at_0']*np.ones(len(q))
    else:
        if 'r0_sphere' in params and 'sigma_sphere' in params:
            I_sphere = compute_spherical_normal_saxs(q,params['r0_sphere'],params['sigma_sphere'])
            I0_sphere = params['I0_sphere']
        else:
            I_sphere = np.zeros(len(q))
            I0_sphere = 0
        if 'r0_pre' in params:
            I_pre = compute_spherical_normal_saxs(q,params['r0_pre'],0)
            I0_pre = params['I0_pre']
        else:
            I_pre = np.zeros(len(q))
            I0_pre = 0
        #if 'q_pk0' in params and 'sigma_pk' in params and 'I0_pk' in params:
        #    I_pk = compute_fcc_peaks()
        #    I0_pk = params['I0_pk']
        #else:
        #    I_pk = np.zeros(len(q))
        #    I0_pk = 0
        I_pk = np.zeros(len(q))
        I0_pk = 0
        return params['I_at_0']*(I0_pre*I_pre + I0_sphere*I_sphere + I0_pk*I_pk)

def compute_spherical_normal_saxs(q,r0,sigma_r):
    """
    Given q, a mean radius r0, and a standard deviation of radius sigma_r,
    compute the saxs spectrum assuming spherical particles 
    with normal size distribution.
    The returned intensity is normalized 
    such that I(q=0) is equal to 1.
    """
    q_zero = (q == 0)
    q_nz = np.invert(q_zero) 
    I = np.zeros(q.shape)
    if sigma_r == 0:
        x = q*r0
        V_r0 = float(4)/3*np.pi*r0**3
        I[q_nz] = V_r0**2 * (3.*(np.sin(x[q_nz])-x[q_nz]*np.cos(x[q_nz]))*x[q_nz]**-3)**2
        I_zero = V_r0**2 
    else:
        dr = sigma_r*0.02
        rmin = np.max([r0-5*sigma_r,dr])
        rmax = r0+5*sigma_r
        I_zero = 0
        for ri in np.arange(rmin,rmax,dr):
            xi = q*ri
            V_ri = float(4)/3*np.pi*ri**3
            # The normal-distributed density of particles with radius r_i:
            rhoi = 1./(np.sqrt(2*np.pi)*sigma_r)*np.exp(-1*(r0-ri)**2/(2*sigma_r**2))
            I_zero += V_ri**2 * rhoi*dr
            I[q_nz] += V_ri**2 * rhoi*dr*(3.*(np.sin(xi[q_nz])-xi[q_nz]*np.cos(xi[q_nz]))*xi[q_nz]**-3)**2
    if any(q_zero):
        I[q_zero] = I_zero
    I = I/I_zero 
    return I

def saxs_Iq4_metrics(q,I):
    """
    This algorithm was developed and 
    originally contributed by Amanda Fournier.    

    From an input spectrum q and I(q),
    compute several properties of the I(q)*q^4 curve.
    This was designed for spectra that are 
    dominated by a dilute form factor term.
    The metrics extracted by this Operation
    were originally intended as an intermediate step
    for estimating size distribution parameters 
    for a population of dilute spherical scatterers.

    Returns a dict of metrics.
    Dict keys and meanings:
    q_at_Iqqqq_min1: q value at first minimum of I*q^4
    I_at_Iqqqq_min1: I value at first minimum of I*q^4
    Iqqqq_min1: I*q^4 value at first minimum of I*q^4
    pIqqqq_qwidth: Focal q-width of polynomial fit to I*q^4 near first minimum of I*q^4 
    pIqqqq_Iqqqqfocus: Focal point of polynomial fit to I*q^4 near first minimum of I*q^4
    pI_qvertex: q value of vertex of polynomial fit to I(q) near first minimum of I*q^4  
    pI_Ivertex: I(q) at vertex of polynomial fit to I(q) near first minimum of I*q^4
    pI_qwidth: Focal q-width of polynomial fit to I(q) near first minimum of I*q^4
    pI_Iforcus: Focal point of polynomial fit to I(q) near first minimum of I*q^4

    TODO: document the algorithm here.
    """
    d = {}
    #if not dI:
    #    # uniform weights
    #    wt = np.ones(q.shape)   
    #else:
    #    # inverse error weights, 1/dI, 
    #    # appropriate if dI represents
    #    # Gaussian uncertainty with sigma=dI
    #    wt = 1./dI
    #######
    # Heuristics step 1: Find the first local max
    # and subsequent local minimum of I*q**4 
    Iqqqq = I*q**4
    # w is the number of adjacent points to consider 
    # when examining the I*q^4 curve for local extrema.
    # A greater value of w filters out smaller extrema.
    w = 10
    idxmax1, idxmin1 = 0,0
    test_range = iter(range(w,len(q)-w-1))
    idx = test_range.next() 
    stop_idx = len(q)-w-1
    while any([idxmax1==0,idxmin1==0]) and idx < stop_idx-1:
        if np.argmax(Iqqqq[idx-w:idx+w+1]) == w and idxmax1 == 0:
            idxmax1 = idx
        if np.argmin(Iqqqq[idx-w:idx+w+1]) == w and idxmin1 == 0 and not idxmax1 == 0:
            idxmin1 = idx
        idx = test_range.next()
    if idxmin1 == 0 or idxmax1 == 0:
        ex_msg = str('unable to find first maximum and minimum of I*q^4 '
        + 'by scanning for local extrema with a window width of {} points'.format(w))
        d['message'] = ex_msg 
        raise Exception(ex_msg)
    #######
    # Heuristics 2: Characterize I*q**4 around idxmin1, 
    # by locally fitting a standardized polynomial.
    idx_around_min1 = (q>0.9*q[idxmin1]) & (q<1.1*q[idxmin1])
    q_min1_mean = np.mean(q[idx_around_min1])
    q_min1_std = np.std(q[idx_around_min1])
    q_min1_s = (q[idx_around_min1]-q_min1_mean)/q_min1_std
    Iqqqq_min1_mean = np.mean(Iqqqq[idx_around_min1])
    Iqqqq_min1_std = np.std(Iqqqq[idx_around_min1])
    Iqqqq_min1_s = (Iqqqq[idx_around_min1]-Iqqqq_min1_mean)/Iqqqq_min1_std
    p_min1 = np.polyfit(q_min1_s,Iqqqq_min1_s,2,None,False,np.ones(len(q_min1_s)),False)
    # polynomial vertex horizontal coord is -b/2a
    qs_at_min1 = -1*p_min1[1]/(2*p_min1[0])
    d['q_at_Iqqqq_min1'] = qs_at_min1*q_min1_std+q_min1_mean
    # polynomial vertex vertical coord is poly(-b/2a)
    Iqqqqs_at_min1 = np.polyval(p_min1,qs_at_min1)
    d['Iqqqq_min1'] = Iqqqqs_at_min1*Iqqqq_min1_std+Iqqqq_min1_mean
    d['I_at_Iqqqq_min1'] = d['Iqqqq_min1']*float(1)/(d['q_at_Iqqqq_min1']**4)
    # The focal width of the parabola is 1/a 
    p_min1_fwidth = float(1)/p_min1[0] 
    d['pIqqqq_qwidth'] = p_min1_fwidth*q_min1_std
    # The focal point is at -b/2a,poly(-b/2a)+1/(4a)
    p_min1_fpoint = Iqqqqs_at_min1+float(1)/(4*p_min1[0])
    d['pIqqqq_Iqqqqfocus'] = p_min1_fpoint*Iqqqq_min1_std+Iqqqq_min1_mean
    #######
    # Heuristics 2b: Characterize I(q) near min1 of I*q^4.
    I_min1_mean = np.mean(I[idx_around_min1])
    I_min1_std = np.std(I[idx_around_min1])
    I_min1_s = (I[idx_around_min1]-I_min1_mean)/I_min1_std
    pI_min1 = np.polyfit(q_min1_s,I_min1_s,2,None,False,np.ones(len(q_min1_s)),False)
    # polynomial vertex horizontal coord is -b/2a
    qs_vertex = -1*pI_min1[1]/(2*pI_min1[0])
    d['pI_qvertex'] = qs_vertex*q_min1_std+q_min1_mean
    # polynomial vertex vertical coord is poly(-b/2a)
    Is_vertex = np.polyval(pI_min1,qs_vertex)
    d['pI_Ivertex'] = Is_vertex*I_min1_std+I_min1_mean
    # The focal width of the parabola is 1/a 
    pI_fwidth = float(1)/pI_min1[0]
    d['pI_qwidth'] = pI_fwidth*q_min1_std
    # The focal point is at -b/2a,poly(-b/2a)+1/(4a)
    pI_fpoint = Is_vertex+float(1)/(4*pI_min1[0])
    d['pI_Ifocus'] = pI_fpoint*I_min1_std+I_min1_mean
    #######
    return d

def saxs_spherical_normal_fit(q,I,I_at_0,method,x_init):
    """
    Fit a saxs spectrum (I(q) vs q) to the theoretical spectrum 
    for dilute spherical nanoparticles with normal size distribution.
    Fit parameters are average radius (r_mean), 
    and standard deviation of radius (sigma_r). 
    The intensity at q=0 (I_at_0) is held fixed.
    The initial estimate of x (x_init) should be 
    an array of the form [r_mean, sigma_r].

    Supported methods (input as strings): 
    full_spectrum_chi2- least squares fit to entire spectrum
    full_spectrum_chi2log- least squares fit to logarithm of entire spectrum
    """
    #try:
    d = {}
    d['message'] = ''
    if method == 'full_spectrum_chi2':
        fit_obj = lambda x: np.sum( (compute_spherical_normal_saxs(q,x[0],x[1]) - I/I_at_0)**2 )
    elif method == 'low_q_chi2':
        fit_obj = lambda x: np.sum( (compute_spherical_normal_saxs(q[:len(q)/2],x[0],x[1]) - I[:len(q)/2]/I_at_0)**2 )
    elif method == 'pearson':
        fit_obj = lambda x: -1*compute_pearson( I, compute_spherical_normal_saxs(q,x[0],x[1]) ) 
    elif method == 'low_q_pearson':
        fit_obj = lambda x: -1*compute_pearson( I[:len(q)/2], compute_spherical_normal_saxs(q[:len(q)/2],x[0],x[1]) ) 
    elif method == 'full_spectrum_chi2log':
        I_nz = np.invert((I==0))
        fit_obj = lambda x: np.sum( (np.log(compute_spherical_normal_saxs(q[I_nz],x[0],x[1])) - np.log(I[I_nz]/I_at_0))**2 )
    elif method == 'low_q_chi2log':
        I_nz = np.invert((I==0)) 
        fit_obj = lambda x: np.sum( (np.log(compute_spherical_normal_saxs(q[I_nz][:len(q)/2],x[0],x[1])) - np.log(I[I_nz][:len(q)/2]/I_at_0))**2 )
    elif method == 'pearson_log':
        I_nz = np.invert((I==0)) 
        fit_obj = lambda x: -1*compute_pearson( np.log(I[I_nz]), np.log(compute_spherical_normal_saxs(q[I_nz],x[0],x[1])) ) 
    elif method == 'low_q_pearson_log':
        I_nz = np.invert((I==0)) 
        fit_obj = lambda x: -1*compute_pearson( np.log(I[I_nz][:len(q)/2]), np.log(compute_spherical_normal_saxs(q[I_nz][:len(q)/2],x[0],x[1])) ) 
    else:
        msg = 'fitting method {} not supported'.format(method)
        raise ValueError(msg)
    res = scipimin(fit_obj,x_init)
    x_opt = res.x
    d['r_mean'] = x_opt[0]
    d['sigma_r'] = x_opt[1]
    return d
    #print('init: {}'.format(fit_obj(x_init)))
    #print('opt: {}'.format(fit_obj(x_opt)))
    #I_opt = I_at_0*compute_spherical_normal_saxs(q,x_opt[0],x_opt[1])
    #I_init = I_at_0*compute_spherical_normal_saxs(q,x_init[0],x_init[1])
    #from matplotlib import pyplot as plt
    #plt.figure(3)
    #plt.semilogy(q,I)
    #plt.semilogy(q,I_init,'r')
    #plt.semilogy(q,I_opt,'g:')
    #plt.show()
    #except Exception as ex:
    #    d['return_code'] = -1
    #    d['message'] = ex.message
    #    return d

def compute_pearson(y1,y2):
    y1mean = np.mean(y1)
    y2mean = np.mean(y2)
    y1std = np.std(y1)
    y2std = np.std(y2)
    return np.sum((y1-y1mean)*(y2-y2mean))/(np.sqrt(np.sum((y1-y1mean)**2))*np.sqrt(np.sum((y2-y2mean)**2)))

def fit_I0(q,I,qmax_fit):
    """
    Find an estimate for I(q=0) by polynomial fitting.
    Inputs q and I are the spectrum,
    qmax_fit is the upper boundary of the fitting region.
    """
    idx_lowq = (q<qmax_fit)
    I_lowq_mean = np.mean(I[idx_lowq])
    I_lowq_std = np.std(I[idx_lowq])
    lowq_mean = np.mean(q[idx_lowq])
    lowq_std = np.std(q[idx_lowq])
    I_lowq_s = (I[idx_lowq]-I_lowq_mean)/I_lowq_std
    lowq_s = (q[idx_lowq]-lowq_mean)/lowq_std
    p_lowq = lowq_spectrum_polyfit(lowq_s,I_lowq_s,-1*lowq_mean/lowq_std,4) 
    I_at_0 = np.polyval(p_lowq,-1*lowq_mean/lowq_std)*I_lowq_std+I_lowq_mean
    return I_at_0

def lowq_spectrum_polyfit(qs,Is,qs_0,order,weights=None):
    """
    Perform a polynomial fitting 
    of the low-q region of the spectrum
    with dI/dq(q=0) constrained to be zero.
    This is performed by forming a Lagrangian 
    from a quadratic cost function 
    and the Lagrange-multiplied constraint function.
    
    TODO: Explicitly document cost function, constraints, Lagrangian.

    Inputs qs and Is are not standardized in this function,
    so they should be standardized beforehand 
    if standardized fitting is desired.
    At the provided constraint point, qs_0, 
    the returned polynomial will have zero slope.
    """
    Ap = np.zeros( (order+1,order+1),dtype=float )
    b = np.zeros(order+1,dtype=float)
    # TODO: vectorize the construction of Ap
    for i in range(0,order):
        for j in range(0,order):
            Ap[i,j] = np.sum( qs**j * qs**i )
        Ap[i,order] = -1*i*qs_0**(i-1)
    for j in range(0,order):
        Ap[order,j] = j*qs_0**(j-1)
        b[j] = np.sum(Is*qs**j)
    p_fit = np.linalg.solve(Ap,b) 
    p_fit = p_fit[:-1]  # throw away Lagrange multiplier term 
    p_fit = p_fit[::-1] # reverse coefs to get np.polyfit format
    #from matplotlib import pyplot as plt
    #plt.figure(3)
    #plt.plot(qs,Is)
    #plt.plot(qs,np.polyval(p_fit,qs))
    #plt.plot(np.arange(qs_0,qs[-1],qs[-1]/100),np.polyval(p_fit,np.arange(qs_0,qs[-1],qs[-1]/100)))
    #plt.plot(qs_0,np.polyval(p_fit,qs_0),'ro')
    #plt.show()
    return p_fit
 

def local_maxima_detector(y):
    """ 
    Finds local maxima in ordered data y.

    :param y: 1d numpy float array
    :return maxima: 1d numpy bool array

    *maxima* is *True* at a local maximum, *False* otherwise.

    This function makes no attempt to reject spurious maxima of any sort.
    """
    length = y.size
    greater_than_follower = np.zeros(length, dtype=bool)
    greater_than_leader = np.zeros(length, dtype=bool)
    greater_than_follower[:-1] = np.greater(y[:-1], y[1:])
    greater_than_leader[1:] = np.greater(y[1:], y[:-1])
    maxima = np.logical_and(greater_than_follower, greater_than_leader)
    # End points
    maxima[0] = greater_than_follower[0]
    maxima[-1] = greater_than_leader[-1]
    return maxima

def local_minima_detector(y):
    """
    Finds local minima in ordered data *y*.

    :param y: 1d numpy float array
    :return minima: 1d numpy bool array

    *minima* is *True* at a local minimum, *False* otherwise.

    This function makes no attempt to reject spurious minima of any sort.
    """
    return local_maxima_detector(-y)




from collections import OrderedDict

import numpy as np
from scipy import interp
from scipy.optimize import minimize

from ...Operation import Operation
from ... import optools
from . import saxstools

class SpectrumParameterization(Operation):
    """
    The algorithm for guessing parameters
    for the size distributions of spherical nanoparticles 
    was developed and originally contributed by Amanda Fournier.    

    This operation uses a SAXS spectrum (I(q) vs. q), 
    along with some profiling information,
    to guess a set of parameters
    for fitting the SAXS spectrum
    to theoretical scattering models. 
    The SAXS model employed here 
    assumes that the spectrum is a sum of the SAXS spectra
    of its constituent particles,
    which can be either precursors (small monodisperse spheres),
    spherical nanoparticles (spheres with a normal size distribution),
    or crystalline arrangements yielding diffraction peaks.
    Any number of populations of these three types can be specified.

    A precursor is modeled by a dilute, monodisperse spherical form factor.
    A spherical nanoparticle population is modeled by
    a discrete sum over a probability distribution
    of dilute, monodisperse spherical form factors. 
    A crystalline arrangement is modeled by a pseudo-Voigt profile.

    Outputs a return code and initial guesses for SAXS model parameters. 
    Also outputs the theoretical result for I(q)
    and a renormalized measured spectrum, for visual comparison.

    This Operation is somewhat robust for noisy data,
    but any preprocessing (background subtraction, smoothing, or other cleaning)
    should be performed beforehand. 
    """

    def __init__(self):
        input_names = ['q', 'I', 'precursor_flag', 'form_flag', 'structure_flag']
        output_names = ['return_code','params','q_I_norm','q_I_guess']
        super(SpectrumParameterization, self).__init__(input_names, output_names)
        self.input_doc['q'] = '1d array of wave vector values in 1/Angstrom units'
        self.input_doc['I'] = '1d array of intensity values I(q)'
        self.input_doc['precursor_flag'] = str('Boolean for whether or not to include '
        + 'monodisperse spherical precursor parameters '
        + '(2 parameters: precursor radius and intensity scaling factor).')
        self.input_doc['form_flag'] = str('Boolean for whether or not to include '
        + 'normally-distributed spherical form factor parameters '
        + '(3 parameters: mean radius, standard deviation of radius, intensity scaling factor).')
        self.input_doc['structure_flag'] = str('Boolean for whether or not to include '
        + 'a set of Pseudo-Voigt fcc diffraction peak parameters '
        + '(3 parameters: principal peak location, peak width, intensity scaling factor).')
        self.output_doc['return_code'] = str('integer indicating whether or not '
        + 'the spectrum was successfully parameterized for the given flags. '
        + 'Possible values: 0=success, 1=error (raised an Exception), '
        + '2=warning (spectrum exhibits unexpected features).')
        self.output_doc['params'] = str('dictionary containing SAXS spectrum parameters ' 
        + 'for populations corresponding to the input flags. ' 
        + 'Dict keys and descriptions: \n' 
        + 'I_at_0: Intensity at q=0, by fitting a polynomial to the low-q region with dI/dq(q=0)=0. \n' 
        + 'r0_pre: precursor term radius (Angstrom). \n'
        + 'I0_pre: precursor intensity scaling factor. \n'
        + 'r0_sphere: mean radius (Angstrom) of sphere population. \n'
        + 'sigma_sphere: standard deviation (Angstroms) of sphere population radii. \n'
        + 'I0_sphere: spherical population intensity scaling factor. \n'
        + 'q_pk0: q value (1/Angstrom) for location of fundamental structure factor peak. \n' 
        + 'sigma_pk: width parameter for Pseudo-Voigt diffraction peak profile. \n' 
        + 'I0_pk: structure factor intensity scaling factor.')
        self.output_doc['q_I_norm'] = 'The input spectrum normalized so that I(q=0) is near 1.'
        self.output_doc['q_I_guess'] = str('n-by-2 array of q and the modeled intensity spectrum '
        + 'for the parameters given in the params output') 
       
        self.input_src['q'] = optools.wf_input
        self.input_src['I'] = optools.wf_input
        self.input_src['precursor_flag'] = optools.wf_input
        self.input_src['form_flag'] = optools.wf_input
        self.input_src['structure_flag'] = optools.wf_input
        self.input_type['q'] = optools.ref_type
        self.input_type['I'] = optools.ref_type
        self.input_type['precursor_flag'] = optools.bool_type
        self.input_type['form_flag'] = optools.bool_type
        self.input_type['structure_flag'] = optools.bool_type
        self.inputs['precursor_flag'] = False 
        self.inputs['form_flag'] = False
        self.inputs['structure_flag'] = False

    def run(self):
        q, I = self.inputs['q'], self.inputs['I']
        # Set return code to 1 (error) by default;
        # if execution finishes, set it to 0
        self.outputs['return_code'] = 1
        p = {} 
        #p = dict.fromkeys(['I_at_0','r0_pre','I0_pre','r0_sphere','sigma_sphere','I0_sphere','q_pk0','sigma_pk','I0_pk']) 
        # Perform a low-q spectrum fit to get I(q=0).
        # Use the lower 5% of the q range.
        I_at_0 = saxstools.fit_I0(q,I,q[0]+float(q[-1]-q[0])*0.1)
        p['I_at_0'] = I_at_0
        if self.inputs['precursor_flag']:
            p['r0_pre'] = 10
        if self.inputs['form_flag']:
            m = saxstools.saxs_Iq4_metrics(q,I)
            width_metric = m['pI_qwidth']/m['q_at_Iqqqq_min1']
            intensity_metric = m['I_at_Iqqqq_min1']/I_at_0
            #######
            #
            # POLYNOMIALS FITTED FOR q0+/-10%,
            # where q0 is the argmin of a parabola
            # that is fit around the first minimum of I*q**4.
            # polynomial for qr0 focus (with x=sigma_r/r0):
            # -8.05459639763x^2 + -0.470989868709x + 4.50108683096
            p_qr0_focus = [-8.05459639763,-0.470989868709,4.50108683096]
            # polynomial for width metric (with x=sigma_r/r0):
            # 3.12889797288x^2 + -0.0645231661487x + 0.0576604958693
            p_w = [3.12889797288,-0.0645231661487,0.0576604958693]
            # polynomial for intensity metric (with x=sigma_r/r0):
            # -1.33327411025x^3 + 0.432533640102x^2 + 0.00263776123775x + -1.27646761062e-05
            p_I = [-1.33327411025,0.432533640102,0.00263776123775,-1.27646761062e-05]
            #
            #######
            # Now find the sigma_r/r0 value that gets the extracted metrics
            # as close as possible to p_I and p_w.
            width_error = lambda x: (np.polyval(p_w,x)-width_metric)**2
            intensity_error = lambda x: (np.polyval(p_I,x)-intensity_metric)**2
            # TODO: make the objective function weight all errors equally
            heuristics_error = lambda x: width_error(x) + intensity_error(x)
            res = minimize(heuristics_error,[0.1],bounds=[(0,0.3)]) 
            if not res.success:
                self.outputs['return_code'] = 2 
                msg = str('[{}] function minimization failed during '
                + 'form factor parameter extraction.'.format(__name__))
                self.outputs['results'] = {'message':msg}
            sigma_over_r = res.x[0]
            qr0_focus = np.polyval(p_qr0_focus,sigma_over_r)
            # qr0_focus = x1  ==>  r0 = x1 / q1
            r0 = qr0_focus/m['q_at_Iqqqq_min1']
            sigma_r = sigma_over_r * r0 
            p['r0_sphere'] = r0
            p['sigma_sphere'] = sigma_r
        if self.inputs['precursor_flag'] and self.inputs['form_flag']:
            p['I0_pre'] = 0.1
            p['I0_sphere'] = 0.9 
        elif self.inputs['precursor_flag']:
            p['I0_pre'] = 1.0 
        elif self.inputs['form_flag']:
            p['I0_sphere'] = 1.0 
        # TODO: process diffraction pk parameters.
        # For now, skip.
        #if self.inputs['structure_flag']:
        #    p['q_pk0'] = 0
        #    p['sigma_pk'] = 0
        #    p['I0_pk'] = 0

        I_guess = saxstools.compute_saxs(q,p)
        I_guess_at_0 = saxstools.compute_saxs(np.array([0]),p)
        q_I_guess = np.array([q,I_guess/I_guess_at_0[0]]).T
        q_I_norm = np.array([q,I/I_at_0]).T
        #### Get logarithmic coef of determination to assess fit quality:
        I_norm_nz = np.invert( (q_I_norm[:,1]==0) )
        #import pdb; pdb.set_trace()
        logI_norm = np.log(q_I_norm[I_norm_nz,1])
        logI_guess = np.log(q_I_guess[I_norm_nz,1])
        sum_logvar = np.sum( (logI_norm-np.mean(logI_norm))**2 )
        sum_logres = np.sum( (logI_guess-logI_norm)**2 ) 
        R2_log = float(1)-float(sum_logres)/sum_logvar
        p['R2_log'] = R2_log
        ####
        self.outputs['q_I_norm'] = q_I_norm 
        self.outputs['q_I_guess'] = q_I_guess 
        self.outputs['params'] = p 
        self.outputs['return_code'] = 0

    def generate_heuristics():
        sigma_over_r = []
        width_metric = []
        intensity_metric = []
        qr0_focus = []
        # TODO: generate heuristics on a 1-d grid of sigma/r
        # instead of the 2-d grid being used here now.
        r0_vals = np.arange(10,41,10,dtype=float)              #Angstrom
        for ir,r0 in zip(range(len(r0_vals)),r0_vals):
            q = np.arange(0.001/r0,float(10)/r0,0.001/r0)       #1/Angstrom
            sigma_r_vals = np.arange(0*r0,0.21*r0,0.01*r0)      #Angstrom
            for isig,sigma_r in zip(range(len(sigma_r_vals)),sigma_r_vals):
                I = compute_spherical_normal_saxs(q,r0,sigma_r) 
                I_at_0 = compute_spherical_normal_saxs(0,r0,sigma_r) 
                d = saxs_spherical_normal_heuristics(q,I)
                sigma_over_r.append(float(sigma_r)/r0)
                qr0_focus.append(d['q_at_Iqqqq_min1']*r0)
                width_metric.append(d['pI_qwidth']/d['q_at_Iqqqq_min1'])
                intensity_metric.append(d['I_at_Iqqqq_min1']/I_at_0)
        # TODO: standardize before fitting, then revert after
        p_qr0_focus = np.polyfit(sigma_over_r,qr0_focus,2,None,False,None,False)
        p_w = np.polyfit(sigma_over_r,width_metric,2,None,False,None,False)
        p_I = np.polyfit(sigma_over_r,intensity_metric,3,None,False,None,False)
        print('polynomial for qr0 focus (with x=sigma_r/r0): {}x^2 + {}x + {}'.format(p_qr0_focus[0],p_qr0_focus[1],p_qr0_focus[2]))
        print('polynomial for width metric (with x=sigma_r/r0): {}x^2 + {}x + {}'.format(p_w[0],p_w[1],p_w[2]))
        print('polynomial for intensity metric (with x=sigma_r/r0): {}x^3 + {}x^2 + {}x + {}'.format(p_I[0],p_I[1],p_I[2],p_I[3]))
        plot = True
        if plot: 
            from matplotlib import pyplot as plt
            plt.figure(1)
            plt.scatter(sigma_over_r,width_metric)
            plt.plot(sigma_over_r,np.polyval(p_w,sigma_over_r))
            plt.figure(2)
            plt.scatter(sigma_over_r,intensity_metric)
            plt.plot(sigma_over_r,np.polyval(p_I,sigma_over_r))
            plt.figure(3)
            plt.scatter(sigma_over_r,qr0_focus)
            plt.plot(sigma_over_r,np.polyval(p_qr0_focus,sigma_over_r))
            plt.figure(4)
            plt.scatter(width_metric,intensity_metric) 
            plt.show()





import numpy as np

from scipy.optimize import curve_fit

from ....Operation import Operation
from .... import optools
from .. import saxstools

class SphericalNormalSpectrumFit(Operation):
    """
    Use a SAXS spectrum (I(q) vs. q),
    for spherical nanoparticles with a normal size distribution,
    to find the mean and standard deviation of particle sizes
    by minimizing an objective function that compares
    the measured spectrum against the theoretical result 
    for those distribution parameters.
    TODO: document the algorithm here.

    Output a return code and diagnostics as needed.
    In case of success, report the size distribution parameters
    and the approximate scattered intensity at q=0.
    Also return the corresponding theoretical result for I(q),
    alongside a renormalized measured spectrum, for visual comparison.

    This Operation is somewhat robust for noisy data,
    but any preprocessing (background subtraction, smoothing, or other cleaning)
    should be performed beforehand. 
    """

    def __init__(self):
        input_names = ['q','I','method','fit_params','ok_flag','dI']
        output_names = ['return_code','results','q_I_norm','q_I_guess']
        super(SphericalNormalSpectrumFit, self).__init__(input_names, output_names)

        self.input_doc['q'] = '1d array of wave vector values in 1/Angstrom units'
        self.input_doc['I'] = '1d array of intensity values I(q)'
        self.input_doc['method'] = str('String indicating choice of optimization method. Options: '
        + '(1) full_spectrum_chi2- sum of difference squared across entire q range. '
        + '(2) full_spectrum_chi2log- sum of difference of logarithm, squared, across entire q range. '
        + '(3) low_q_chi2- sum of difference squared in only the lowest half of measured q range. '
        + '(4) low_q_chi2log- sum of difference of logarithm, squared, in lowest half of measured q range. '
        + '(5) pearson- pearson correlation between measured and modeled spectra. '
        + '(6) pearson_log- pearson correlation between logarithms of measured and modeled spectra.'
        + '(7) low_q_pearson- pearson correlation between measured and modeled spectra. '
        + '(8) low_q_pearson_log- pearson correlation between logarithms of measured and modeled spectra.') 
        self.input_doc['fit_params'] = str('dict of inputs for fitting, including initial guess parameters and I(q=0). '
        + 'Should be a dict with the following keys (as strings): "r_mean", "sigma_r", "I_at_0". '
        + 'Expected units of the values, respectively: Angstrom, Angstrom, counts.')
        self.input_doc['ok_flag'] = str('Boolean indicating whether or not this optimization should be attempted. '
        + 'This is useful for skipping the optimization based on the results of previous operations.')
        self.input_doc['dI'] = '1d array of error estimates of intensity values (optional- input None to ignore)'
        self.output_doc['return_code'] = str('integer indicating whether or not the spectrum was found to be fittable. '
        + 'Possible values: 0=success, 1=error, 2=warning')
        self.output_doc['results'] = str('dictionary containing results '
        + 'for mean size, standard deviation, and I(q=0). '
        + 'Dict keys (strings): "r_mean", "sigma_r", "I_at_0". '
        + 'Units: Angstrom, Angstrom, counts.')
        self.output_doc['q_I_norm'] = 'The input q,I(q) spectrum normalized so that I(q=0) is near 1'
        self.output_doc['q_I_guess'] = str('n-by-2 array of q and the theoretical intensity spectrum '
        + 'for the r_mean, sigma_r, and I_at_0 values in the results output') 

        self.input_src['q'] = optools.wf_input
        self.input_src['I'] = optools.wf_input
        self.input_src['method'] = optools.text_input
        self.input_src['fit_params'] = optools.wf_input
        self.input_src['ok_flag'] = optools.wf_input
        self.input_src['dI'] = optools.no_input
        self.input_type['q'] = optools.ref_type
        self.input_type['I'] = optools.ref_type
        self.input_type['method'] = optools.str_type
        self.input_type['fit_params'] = optools.ref_type
        self.input_type['ok_flag'] = optools.ref_type
        # TODO: establish a good default here.
        #self.inputs['method'] = 'full_spectrum_chi2log' 

    def run(self):
        # Set return code to 1 (error) by default;
        # if execution finishes, set it to 0
        self.outputs['return_code'] = 1
        if self.inputs['ok_flag']:
            q, I, dI = self.inputs['q'], self.inputs['I'], self.inputs['dI']
            m = self.inputs['method']
            params = self.inputs['fit_params']
            x_init = [params['r_mean'],params['sigma_r']]
            try:
                d_opt = saxstools.saxs_spherical_normal_fit(q,I,params['I_at_0'],m,x_init,dI)
            except Exception as ex:
                self.outputs['return_code'] = 1
                msg = str('[{}] optimization by {} failed '
                + 'while attempting to fit spectrum. message: '
                + '{}'.format(__name__,m,ex.message))
                self.outputs['results'] = {'r_mean':0,'sigma_r':0,'I_at_0':0,'message':ex.message} 
                raise ex
            r0 = d_opt['r_mean']
            sigma_r = d_opt['sigma_r']
            I_at_0 = d_opt['I_at_0']
            self.outputs['results'] = {'r_mean':r0,'sigma_r':sigma_r,'I_at_0':I_at_0} 
            self.outputs['q_I_norm'] = q_I_norm 
            self.outputs['q_I_guess'] = q_I_guess 
            self.outputs['return_code'] = 1
        else:        
            self.outputs['results'] = {'r_mean':0,'sigma_r':0,'I_at_0':0,'message':'ok_flag was set to False- optimization skipped'} 




import numpy as np
from scipy import interp
from scipy.optimize import minimize

from ...Operation import Operation
from ... import optools
from . import saxstools

class SpectrumProfiler(Operation):
    """
    This operation profiles a SAXS spectrum (I(q) vs. q)
    to determine some characteristics of the sample.
    Based on some measures of the overall shape of the spectrum,
    the spectrum is tested for scattering from precursors 
    (approximated as small dilute monodisperse spheres),
    scattering from spherical nanoparticles 
    (dilute spheres with a normal size distribution),
    diffraction peaks
    (Voigt-like profiles due to crystalline arrangements),
    or some combination of the three.
    TODO: document algorithm here.

    Output a return code
    and a dictionary of the results.

    This Operation is somewhat robust for noisy data,
    but any preprocessing (background subtraction, smoothing, or other cleaning)
    should be performed beforehand. 
    """

    def __init__(self):
        input_names = ['q', 'I', 'dI']
        output_names = ['return_code','precursor_flag','form_flag','structure_flag','metrics']
        super(SpectrumProfiler, self).__init__(input_names, output_names)
        self.input_doc['q'] = '1d array of wave vector values in 1/Angstrom units'
        self.input_doc['I'] = '1d array of intensity values I(q)'
        self.input_doc['dI'] = 'optional- 1d array of intensity uncertainty values I(q)'
        self.output_doc['return_code'] = str('integer indicating '
        + 'whether or not the profiling finished. '
        + 'Possible values: 0=success, 1=error (raised an Exception), '
        + '2=warning (input data has unexpected characteristics).')
        self.output_doc['precursor_flag'] = 'Boolean indicating presence of precursor terms. '
        self.output_doc['form_flag'] = 'Boolean indicating presence of form factor terms. '
        self.output_doc['structure_flag'] = 'Boolean indicating presence of structure factor terms. '
        self.output_doc['metrics'] = str('dict profiling the input spectrum '
        + 'with various numerical measures of the overall spectral profile. '
        + 'Dict keys and descriptions: \n'
        + 'bin_strengths: Integrated intensity in 10 evenly spaced q-bins. \n'
        + 'Imax_over_Imean: maximum intensity divided by mean intensity. \n'
        + 'Imax_over_Ilowq: maximum intensity divided by average intensity over lowest 10% of q domain. \n'
        + 'q_Imax: q value of the maximum intensity. \n'
        + 'q_corr: Pearson correlation between I(q) and q. \n'
        + 'qsquared_corr: Pearson correlation between I(q) and q^2. \n'
        + 'cos2q_corr: Pearson correlation between I(q) and (cos(q))^2. \n'
        + 'cosq_corr: Pearson correlation between I(q) and cos(q). \n'
        + 'invq_corr: Pearson correlation between I(q) and 1/q. \n'
        + 'q_logcorr: Same as q_corr, but with log(I(q)). \n'
        + 'qsquared_logcorr: Same as qsquared_corr, but with log(I(q)). \n'
        + 'cos2q_logcorr: Same as cos2q_corr, but with log(I(q)). \n'
        + 'cosq_logcorr: Same as cosq_corr, but with log(I(q)). \n'
        + 'invq_logcorr: Same as invq_corr, but with log(I(q)).')
        self.input_src['q'] = optools.wf_input
        self.input_src['I'] = optools.wf_input
        self.input_src['dI'] = optools.no_input
        self.input_type['q'] = optools.ref_type
        self.input_type['I'] = optools.ref_type

    def run(self):
        q, I = self.inputs['q'], self.inputs['I']
        dI = self.inputs['dI']
        # Set return code to 1 (error) by default;
        # if execution finishes, set it to 0
        self.outputs['return_code'] = 1
	    ###check correlations of input vs. some known trends...
        n_q = len(q)
	    # correlations on intensity
        q_corr = saxstools.compute_pearson(I,np.linspace(0,1,n_q))
        qsquared_corr = saxstools.compute_pearson(I,np.linspace(0,1,n_q)**2)
        cosq_corr = saxstools.compute_pearson(I,np.cos(q*np.pi/(2*q[-1])))
        cos2q_corr = saxstools.compute_pearson(I,np.cos(q*np.pi/(2*q[-1]))**2)
        invq_corr = saxstools.compute_pearson(I,q**-1)
        # correlations on log intensity
        idx_nz = np.invert(I == 0)
        q_logcorr = saxstools.compute_pearson( np.log(I[idx_nz]) , np.linspace(0,1,n_q)[idx_nz] )
        qsquared_logcorr = saxstools.compute_pearson( np.log(I[idx_nz]) , (np.linspace(0,1,n_q)[idx_nz])**2 )
        cosq_logcorr = saxstools.compute_pearson( np.log(I[idx_nz]) , np.cos(q*np.pi/(2*q[-1]))[idx_nz] )
        cos2q_logcorr = saxstools.compute_pearson( np.log(I[idx_nz]) , (np.cos(q*np.pi/(2*q[-1]))[idx_nz])**2 )
        invq_logcorr = saxstools.compute_pearson( np.log(I[idx_nz]) , q[idx_nz]**-1 )
        #invq4_corr = saxstools.compute_pearson(I,q**-4)
        #I0 = I[np.min(np.argwhere(I>0))]
        idxmax = np.argmax(I)
        Imax = I[idxmax] 
        q_Imax = q[idxmax]
        Imean = np.mean(I)
        Imax_over_Imean = float(Imax)/float(Imean)
        ###compute bin-integrated spectral intensities.
        ### TODO: use these somehow
        qmin, qmax = q[0], q[-1]
        qrange = qmax-qmin
        bin_strengths = np.zeros(10)
        for i in range(10):
            qmini, qmaxi = qmin+i*qrange/10,qmin+(i+1)*qrange/10
            idxi = list((q>=qmini) & (q<qmaxi))
            idxi_shift = [False]+idxi[:-1]
            dqi = q[ idxi ] - q[ idxi_shift ]
            Ii = I[ idxi ]
            bin_strengths[i] = np.sum(Ii * dqi) / (qmaxi-qmini)
        Imax_over_Ilowq = float(Imax)/bin_strengths[0]
        # Flagging form factor: These tend to correlate with 1/q
        form_flag = invq_logcorr > 0.7 or invq_corr > 0.3 
        # Flagging precursors: Only flag if there is also a form term.
        # (Assume fitting precursors alone is boring).
        precursor_flag = form_flag and ( any([cosq_corr > 0.6, cos2q_corr > 0.6])
                        or all([cosq_logcorr>0.4,cos2q_logcorr>0.4,cosq_corr>0.5,cos2q_corr>0.5]) )
        # Flagging structure: look for large values away from q=0.
        # TODO: more detailed structure factor evaluation. 
        structure_flag = Imax_over_Ilowq > 1.2 and Imax_over_Imean > 4 and q_Imax > 0.05
        #######
        # Any good/bad filters should be added here.
        ok_flag = True
        flag_msg = ''
        if q_corr > 0.2:
            ok_flag = False
            flag_msg = str('This algorithm expects spectra that tend to decrease in q. '
            + 'The input spectrum seems to be overall increasing in q.')
        #if (...) 
        #    ok_flag = False
        #    flag_msg = 'This spectrum seems mostly flat.'
        #elif (...):
        #    ok_flag = False
        #    flag_msg = ''
        #######
        d_r = {}
        d_r.update({'bin_strengths':bin_strengths})
        d_r.update({'Imax_over_Imean':Imax_over_Imean})
        d_r.update({'Imax_over_Ilowq':Imax_over_Ilowq})
        d_r.update({'q_Imax':q_Imax})
        d_r.update({'q_logcorr':q_logcorr})
        d_r.update({'qsquared_logcorr':qsquared_logcorr})
        d_r.update({'cos2q_logcorr':cos2q_logcorr})
        d_r.update({'cosq_logcorr':cosq_logcorr})
        d_r.update({'invq_logcorr':invq_logcorr})
        d_r.update({'q_corr':q_corr})
        d_r.update({'qsquared_corr':qsquared_corr})
        d_r.update({'cos2q_corr':cos2q_corr})
        d_r.update({'cosq_corr':cosq_corr})
        d_r.update({'invq_corr':invq_corr})
        self.outputs['metrics'] = d_r 
        if not ok_flag:
            d_r['WARNING_MESSAGE'] = flag_msg
            self.outputs['precursor_flag'] = False 
            self.outputs['form_flag'] = False 
            self.outputs['structure_flag'] = False
            self.outputs['return_code'] = 1
        else:
            self.outputs['precursor_flag'] = precursor_flag
            self.outputs['form_flag'] = form_flag
            self.outputs['structure_flag'] = structure_flag
            self.outputs['return_code'] = 0

        #low_q_idxs = (q < q[0]+0.1*(q[-1]-q[0]))
        #high_q_idxs = (q > q[0]+0.9*(q[-1]-q[0]))
        #n_low_q = np.sum(np.array(low_q_idxs))
        #n_high_q = np.sum(np.array(high_q_idxs))
        # If the max intensity is outside the low-q region, throw flag
        #if not np.argmax(I) in range(2*n_low_q):
        #    ok_flag = False
        #    flag_msg = str('The maximum intensity '
        #    + '({} at q={}) '.format(np.max(I),q[np.argmax(I)])
        #    + 'is outside the lower 20% of the q-range '
        #    + '({} to {}).'.format(q[0],q[2*n_low_q]))
        # If there is a huge maximum somewhere, throw flag
        #elif np.max(I) > 10*np.mean(I[:2*n_low_q]):
        #    ok_flag = False
        #    flag_msg = str('The maximum intensity '
        #    + '({} at q={}) '.format(np.max(I),q[np.argmax(I)])
        #    + 'is greater than 10x the mean intensity '
        #    + 'of the lower 20% of the q-range '
        #    + '({} from {} to {}).'
        #    .format(np.mean(I[low_q_idxs]),q[0],q[n_low_q]))
        # If high-q intensity not significantly smaller than low-q, throw flag
        #elif not np.sum(I[:n_low_q]) > 10*np.sum(I[-1*n_high_q:]):
        #    ok_flag = False
        #    flag_msg = str('The lower 10% of the q-range '
        #    + '({} to {}) '.format(q[0],q[n_low_q])
        #    + 'has less than 10 times the intensity '
        #    + 'of the upper 10% of the q-range '
        #    + '({} to {}).'.format(q[-1*n_high_q],q[-1]))


